/*
    FEARLUS model0-5-1-5: BatchSwarm.m
    Copyright (C) 1999-2004  Macaulay Institute

    This file is part of FEARLUS model0-5-1-5, an agent-based model of land use
    change.

    FEARLUS model0-5-1-5 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-5-1-5 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the BatchSwarm class.

*/

#import "BatchSwarm.h"
#import "ModelSwarm.h"
#import "Environment.h"
#import "LandAllocator.h"
#import "FearlusArguments.h"
#import "Parameter.h"
#import "LandUse.h"
#import "MiscFunc.h"		// getUsableFileNameFrom:withSuffix:
#import "SubPopulation.h"
#import "Reporter.h"

@implementation BatchSwarm

/*

create

Create a BatchSwarm object.

*/

+create: (id)z {
  BatchSwarm *o;

  o = [super create: z];
  o->writeHistoryFrequency = 500;
  o->dumpHistory = YES;
  o->saveDumpedHistory = NO;
  o->historyFileName = (char *)malloc(1024);
  sprintf(o->historyFileName, [arguments getHistoryFile]);

  return(o);
}

/*

buildReporter

Build the reporter object, and its schedule

*/

-buildReporter {
  reporter = [Reporter createBegin: self];
  [reporter setParameters: parameter];
  [reporter setModelSwarm: modelSwarm];
  reporter = [reporter createEnd];
  reporterSchedule = [Schedule createBegin: self];
  [reporterSchedule setRepeatInterval: 1];
  reporterSchedule = [reporterSchedule createEnd];
  [reporterSchedule at: 0 createActionTo: reporter message: M(report)];
  [reporterSchedule activateIn: self];

  return self;
}

/*

buildObjects

Build the objects needed for the class

*/

-buildObjects {
  struct stat modelfile;
  double r, g, b;
  int i, initZoom, maxEnvDim;

  [super buildObjects];
  // Build the parameter object
  parameter = [Parameter createBegin: self];
  if([MiscFunc fileExists: [arguments getModelFile]]) {
				// Test to see if a model file exists before
                                // loading it.
    if([arguments debugObserver]) {
      printf("o: Loading model parameters from file %s\n",
	     [arguments getModelFile]);
    }
    [parameter loadFromFileNamed: [arguments getModelFile]];
    dumpHistory = [parameter dumpHistory];
    writeHistoryFrequency = [parameter maxHistoryLength];
  }
  else {
    fprintf(stderr, "Model parameter file %s not found -- using default values"
	    " for parameters\n", [arguments getModelFile]);
    [parameter setDumpHistory: dumpHistory];
    [parameter setMaxHistoryLength: writeHistoryFrequency];
  }
  if([arguments seedHasBeenSpecified]) {
    [parameter setSpecifiedSeed: seed];
  }
  parameter = [parameter createEnd];
  [parameter finalise];
  // Build the model swarm -- must wait until parameter finalised.
  modelSwarm = [ModelSwarm createBegin: self];
  [modelSwarm setParameters: parameter];
  modelSwarm = [modelSwarm createEnd];

  if([arguments useReporter]) {
    reporter = [Reporter createBegin: self];
    [reporter setParameters: parameter];
    [reporter setModelSwarm: modelSwarm];
    reporter = [reporter createEnd];
  }

  // Now I was reading in the library about control panels, I think they are
  // called activity controls, and am I sure that I don't want to use one of
  // those?

  [parameter setDumpHistory: dumpHistory];
  [parameter setMaxHistoryLength: writeHistoryFrequency];
  [modelSwarm buildObjects];

  return(self);
}


/*

buildActions

Over-ride the Swarm buildActions method to introduce the scheduling required
for this class.

*/

-buildActions {
  [super buildActions];
  [modelSwarm buildActions];
  historySchedule = [Schedule createBegin: self];
  [historySchedule setRepeatInterval: writeHistoryFrequency];
  historySchedule = [historySchedule createEnd];
  [historySchedule at: 0 createActionTo: self
		   message: M(writeHistoryToFile:) : (id)historyFileName];

  initialSchedule = [Schedule createBegin: self];
  initialSchedule = [initialSchedule createEnd];
  if([arguments useReporter]) {
    [initialSchedule at: 0 createActionTo: self message: M(buildReporter)];
  }
  if(saveDumpedHistory) {
    [initialSchedule at: writeHistoryFrequency - 1
		     createActionTo: historySchedule
		     message: M(activateIn:) : self];
  }
  if(![parameter infiniteTime]) {
    [initialSchedule at: [parameter maximumYear]
		     createActionTo: self
		     message: M(stop)];
  }
  else {
    fprintf(stderr, "%s -- Batch mode and infinite time -- not a good "
	    "combination!\n");
    exit(1);
  }

  return(self);
}

/*

activateIn

Over-ride the Swarm activateIn method to activate the schedules

*/

-activateIn: (id)swarmContext {
  [super activateIn: swarmContext];
  [modelSwarm activateIn: self];	// model schedule must come first
  // (This means it will get done first, so the display always occurs after
  // the model cycle)
  [initialSchedule activateIn: self];
  return([self getSwarmActivity]);
}

/*

go

Following the Mousetrap BatchSwarm example, this object implements a go method
that is inherited in the ObserverSwarm from the GUISwarm object. We have to
implement our own, however.

*/

-go {
  [[self getActivity] run];             // Run it!
 
  return [[self getActivity] getStatus];
}

/*

stop

Stop the simulation (again, following the Mousetrap BatchSwarm example).

*/

-stop {
  [getTopLevelActivity() terminate];
 
  return self;
}

/*

setSpecifiedSeed:

If the user specified a seed on the command line, then it should be recorded
in the history file. So it needs to be passed in from main.

*/

-(void)setSpecifiedSeed: (unsigned)value {
  seed = value;
}

/*

writeHistory

Call the modelSwarm to get it to write its history to a file that has been
created. This should be made available as a button so it can be called at any
time. If this method has a return type of void, it cannot be made available as
a button. (You get an abort when you click it.)

If the user has specified on the command line that they want a history file
created, then clicking the button will have no effect. A full history will be
written when they quit the application.

*/

-writeHistory {
  if([arguments historyFileHasBeenSpecified]) {
    fprintf(stderr, "History will be written when you quit\n");
  }
  else {
    [self writeHistoryToFile: historyFileName];
  }
  return self;
}

/*

writeHistoryToFile

The method that actually writes the history to a file. Empty file names will
not be written to. This method enforces a .txt suffix for XL. If the file 
exists, it will insert the minimum three digit number before the .txt suffix
required to make the filename unique.

*/

-(void)writeHistoryToFile: (const char *)fileName {
  char *realFileName;
  FILE *fp;

  realFileName = [MiscFunc getUsableFileNameFrom: fileName withSuffix: ".txt"];
  if(realFileName == NULL) {
    fprintf(stderr, "%s -- cannot write history because couldn't create "
	    "datafile\n", sel_get_name(_cmd));
    return;
  }
  // Now we actually write the file
  fp = fopen(realFileName, "w");
  if(!fp) {
    fprintf(stderr, "Cannot create datafile %s\n", realFileName);
    perror(realFileName);
  }
  else {
    printf("Saving data to file: %s\n", realFileName);
    [modelSwarm writeHistory: fp];
    fclose(fp);
  }
  free(realFileName);
}

/*

drop

Over-ride the Swarm drop method to make sure all items created by this object
are dropped

*/

-(void)drop {
  [modelSwarm drop];
  free(historyFileName);
  [super drop];
}

@end
