/*
    FEARLUS model0-5-1-5: ImitativeOptimumStrategy.m
    Copyright (C) 1999-2004  Macaulay Institute

    This file is part of FEARLUS model0-5-1-5, an agent-based model of land use
    change.

    FEARLUS model0-5-1-5 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-5-1-5 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the ImitativeOptimumStrategy object. 

*/

#import "ImitativeOptimumStrategy.h"
#import "SelectUseBucket.h"
#import "LandUse.h"
#import "LandParcel.h"
#import "LandManager.h"
#import "Environment.h"
#import "Parameter.h"
#import "FearlusArguments.h"
#import "BitString.h"
#import "MiscFunc.h"

@implementation ImitativeOptimumStrategy
/*

create:withManager:andParameters:

Create the strategy, setting the manager and parameters.

*/

+(id <ImitativeStrategy, HistoricalStrategy>)create: aZone
	   withManager: (LandManager *)lmgr
	 andParameters: (Parameter *)p {
  ImitativeOptimumStrategy *obj;

  obj = [super create: aZone];
  obj->lm = lmgr;
  obj->parameter = p;
  return obj;
}

/*

decideLandUseForParcel:

Decide the land use for the land parcel, using an imitative optimum strategy.
A formula is used to calculate the expected yield for the decision parcel given
the yield, land use and bit string of the neighbouring parcel. The expected
yield is consistent, so once an exemplar has been found for a land use, then
any further parcels with that land use are ignored. The buckets will then
contain the expected yield obtained for each land use observed for the decision
parcel. A random selection is made among those with the highest score. The
expected yield calculation will not cope with don't care bits (when present,
these land managers will no longer be imitative *optimisers*).

This method does make use of neighbourhood weighting and neighbourhood noise.
Neighbourhood noise will disrupt the bit string of the neighbouring parcels
belonging to other land managers. The effect will be some miscalculation of the
expected yield. Neighbourhood weighting can be used to mitigate these effects.
A lower neighbourhood weighting will mean land managers are less likely to use
land uses that they do not already use.

The land manager's own land parcels are always inspected first. Thus any
neighbour's parcel with the same land use will be ignored. This will further
mitigate the effects of neighbourhood noise.

*/

-(LandUse *)decideLandUseForParcel: (LandParcel *)lp {
  SelectUseBucket *bucket;
  id parcelIndex, parcelList, tempZone;
  LandParcel *myParcel, *neighbouringParcel;
  LandManager *neighbour;
  LandUse *lu;
  int i;
  unsigned yield;
  unsigned match;
  BitString *bs;
  double getExpectedYield(unsigned yield, unsigned antimatch,
			     unsigned lubits) {
    double eyield;
    int i;

    eyield = 0.0;
    if([arguments debugInnerCoreMethods]) {
      printf("i: Calculation of Expected Yield:\n\tYield = %u, antimatch = %u,"
	     " land use bits = %u\n", yield, antimatch, lubits);
      printf("\tYield        No. of ways             Out of   Probability\n");
    }
    for(i = 0; i <= (int)yield; i++) {
      eyield += ((double)((int)yield + (int)antimatch - (2 * i))
		 * (double)[MiscFunc combineN: (int)antimatch r: i]
		 * (double)[MiscFunc combineN: (int)(lubits - antimatch)
				     r: (int)yield - i]);
      if([arguments debugInnerCoreMethods]) {
	unsigned long long ways, total;

	ways = [MiscFunc combineN: (int)antimatch r: i]
	  * [MiscFunc combineN: (int)(lubits - antimatch) r: (int)yield - i];
	total = [MiscFunc combineN: (int)lubits r: (int)yield];
	printf("\t% 5d % 18lld % 18lld % 13g\n",
	       ((int)yield + (int)antimatch - (2 * i)), ways, total,
	       (total > 0 ? (double)ways / (double)total : 0.0));
      }
    }
    eyield /= (double)[MiscFunc combineN: (int)lubits r: (int)yield];
    return eyield;
  }

  tempZone = [Zone create: scratchZone];
  bucket = [SelectUseBucket create: tempZone
			    withParameters: parameter
			    andLandUses: [[lp getEnvironment] getLandUses]];
  // Loop through land parcels owned by the land manager
  if([arguments debugCoreMethods]) {
    printf("S: %s: Using imitative optimum strategy of land manager %u to "
	   "determine land use for LP = land parcel %u at (%d, %d)\n",
	   sel_get_name(_cmd), [lm getPIN], [lp getPIN], [lp getX], [lp getY]);
  }
  for(parcelIndex = [[lm getLandParcels] begin: tempZone], [parcelIndex next];
      [parcelIndex getLoc] == Member;
      [parcelIndex next]) {
    myParcel = [parcelIndex get];
    yield = [myParcel getLastYield];
    match = [[lp getAttributes] getMatchWith: [myParcel getAttributes]];
    if([bucket getScoreInBucketOfLandUse: [myParcel getLandUse]] == 0.0) {
      double score;

      score = getExpectedYield(yield,
			       [parameter parcelBitStringSize] - match,
			       [parameter useBitStringSize]);
      [bucket addScore: score toBucketForLandUse: [myParcel getLandUse]];
      if([arguments debugCoreMethods]) {
	printf("S: %s: Found new land use %u at land parcel %u at (%d, %d) "
	       " with antimatch to LP of %u and yield %u. Expected yield "
	       " (= score for LU) = %g\n", sel_get_name(_cmd),
	       [[myParcel getLandUse] getPIN], [myParcel getPIN],
	       [myParcel getX], [myParcel getY], 
	       [parameter parcelBitStringSize] - match, yield, score);
      }
    }
    else {
      if([arguments debugCoreMethods]) {
	printf("S: %s: Already got expected yield (%g) for land use %u of "
	       "land parcel %u at (%d, %d). Recalculation of expected yield "
	       "gives %g\n", sel_get_name(_cmd),
	       [bucket getScoreInBucketOfLandUse: [myParcel getLandUse]],
	       [[myParcel getLandUse] getPIN], [myParcel getPIN],
	       [myParcel getX], [myParcel getY], 
	       getExpectedYield(yield,
				[parameter parcelBitStringSize] - match,
				[parameter useBitStringSize]));
      }
    }
  }
  [parcelIndex drop];
  // Loop through the neighbours, and through their land parcels
  bs = [BitString create: tempZone
		  setBitCount: [parameter parcelBitStringSize]];
  for(i = 0; neighbour = [lm getNextNeighbour: i]; i |= 1) {
    if([arguments debugCoreMethods]) {
      printf("S: %s: Adding scores for land uses of land parcels belonging"
	     " to land manager %u\n", sel_get_name(_cmd), [neighbour getPIN]);
    }
    for(parcelIndex = [[neighbour getLandParcels] begin: tempZone],
	  [parcelIndex next];
	[parcelIndex getLoc] == Member;
	[parcelIndex next]) {
      neighbouringParcel = [parcelIndex get];
      [bs setToCopyOf: [neighbouringParcel getAttributes]];
      [bs mutateWithProbability: [lm getNeighbourNoise]];
      match = [[lp getAttributes] getMatchWith: bs];
      yield = [neighbouringParcel getLastYield];
      if([bucket getScoreInBucketOfLandUse: [neighbouringParcel getLandUse]]
	 == 0.0) {
	double score;

	score = getExpectedYield(yield,
				 [parameter parcelBitStringSize] - match,
				 [parameter useBitStringSize]);
	[bucket addScore: score * [lm getNeighbourWeight] 
		toBucketForLandUse: [neighbouringParcel getLandUse]];

	if([arguments debugCoreMethods]) {
	  printf("S: %s: Found new land use %u at neighbour's land parcel %u "
		 "at (%d, %d) with antimatch to LP of %u and yield %u. "
		 "Expected yield = %g. Score added = %g\n", sel_get_name(_cmd),
		 [[neighbouringParcel getLandUse] getPIN], 
		 [neighbouringParcel getPIN], [neighbouringParcel getX],
		 [neighbouringParcel getY],
		 [parameter parcelBitStringSize] - match, yield, score,
		 score * [lm getNeighbourWeight]);
	}
      }
      else {
	if([arguments debugCoreMethods]) {
	  double score;

	  score = getExpectedYield(yield,
				   [parameter parcelBitStringSize] - match,
				   [parameter useBitStringSize]);
	  printf("S: %s: Already got expected yield (%g) for land use %u of "
		 "land parcel %u at (%d, %d). Recalculation of expected yield "
		 "gives %g -- with neighbourhood weighting: %g\n",
		 sel_get_name(_cmd),
		 [bucket getScoreInBucketOfLandUse: 
			   [neighbouringParcel getLandUse]],
		 [[neighbouringParcel getLandUse] getPIN],
		 [neighbouringParcel getPIN],
		 [neighbouringParcel getX], [neighbouringParcel getY], 
		 score, score * [lm getNeighbourWeight]);
	}
      }
    }
    [parcelIndex drop];
  }
  lu = [bucket getRandomOptimumChoice];
  [bucket drop];
  [bs drop];
  [tempZone drop];
  return lu;
}

@end
