/*
    FEARLUS model0-5-1-5: Parameter.h
    Copyright (C) 1999-2004  Macaulay Institute

    This file is part of FEARLUS model0-5-1-5, an agent-based model of land use
    change.

    FEARLUS model0-5-1-5 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-5-1-5 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


This is the interface for the parameters class. It contains all the model
parameters. Objects requiring to know the value of a particular parameter
will need to have a method to be passed the parameter object in. They can then
store a pointer to the parameter object in their instance variables. This is
a way of avoiding objects having to point back to the modelSwarm, which maybe
isn't such a good idea.

The SubPopulation array makes things a little more tricky. There should be a
method which displays an array of buttons, with one button for each
sub-population. Clicking on a button should result in that sub-population's
probemap being displayed. This method should appear in the probemap for the
parameter class.

*/

#import <objectbase/SwarmObject.h>
#import <objectbase/DefaultProbeMap.h>
#import <stdio.h>
#import <objc/objc-api.h>	// sel_get_name
#import <tkobjc/ButtonPanel.h>
#import <collections.h>
#import <simtools.h>
#import "Clumper.h"

@class SubPopulation, NumberArray, StrategyColourKey;

/*

In case you worry about this again:

There's no problem with the char ptr variables having no memory allocated to
them. All settings, whether by being loaded from a file using the objectloader
or on the GUI, are done through probes. The probes use strdup to copy the
string, so memory gets allocated for the char pointer. Strictly, this means
we should free all charptr instance variables when dropping the object, unless
the drop method does this for us already. See VarProbe.m -- setData method
for proof of use of strdup.

*/

@interface Parameter: SwarmObject {
  // Instance variables with a causal influence (model parameters)
  // 	Parameters that apply to the environment
  char *environmentType;	// The type of the environment
  unsigned neighbourhoodRadius;	// The "radius" in which to look for neighbours
  
  /* deprecated parameters */
  int neighbourhoodIsMoore;	// 0 for a von Neumann nbrhood 1 for Moore
  int environmentIsBounded;	// 0 for toroidal environment 1 for bounded
  /* end deprecated parameters */
  unsigned climateBSSize;	// Size of the climate bit string
  NumberArray *climateToggleProbs;
  unsigned economyBSSize;	// Size of the economy bit string
  NumberArray *economyToggleProbs;
  unsigned landParcelBSSize;
				// Size of the land parcel bit string
  unsigned nLandUse;		// Number of land use types
  double pLandUseDontCare;
				// Probability of "Don't Care" bit in land use
				// bit string
  char *clumping;		// String of clumping class and settings
  id <Clumper> clumperobj;	// The actual clumping class object
  unsigned nLandParcelClump;	// (Deprecated)
				// Number of cycles of land parcel clumping CA
  unsigned int envXSize;	// Number of horizontal cells in environment
  unsigned int envYSize;	// Number of vertical cells in environment
  char *landParcelFile;		// File to load land parcels from
  char *landUseFile;		// File to load land uses from
  char *climateFile;		// File to load climate from
  char *climateToggleProbFile;
  char *economyFile;		// File to load economy from
  char *economyToggleProbFile;
  BOOL useLandParcelFile;	// Whether or not to use the land parcel file
  BOOL useLandUseFile;		// Whether or not to use the land use file
  BOOL useClimateFile;		// Whether or not to use the climate file
  BOOL useEconomyFile;		// Whether or not to use the economy file
  unsigned maxYear;		// The year at which to terminate the sim
  BOOL infiniteTime;		// Whether or not to use the maximum year

  // 	Parameters that apply to the land allocation body
  unsigned nSubPops;		// Number of sub-populations
  id <Array> subPops;		// Array of sub-populations
  char *subPopFile;		// File to load sub-populations from
  double strategyChangeUnit;	// Unit of increase or decrease of strategy
  double neighbourNoiseMax;   	// Maximum proportion of neighbouring land
				// parcel bit string noise
  double neighbourNoiseMin;	// Minimum proportion of neighbouring land
				// parcel bit string noise
  double breakEvenThreshold;
				// Yield threshold for breaking even on a land
                                // parcel.
  double landParcelPrice;	// Fixed price for land parcels
  unsigned nInitXParcels;	// The number of initial Land parcels to
   				// allocate to Land Managers on the X axis
  unsigned nInitYParcels;	// The number of initial Land parcels to
  				// allocate to Land Managers on the Y axis
  unsigned maxMemorySize;	/* The maximum memory size any Land Manager
				   can possibly be allocated -- computed from 
				   the SubPopulations array */
  //	Parameters for recording history
  BOOL dumpHistory;		// Flag for whether or not to dump the history
  unsigned maxHistoryLength;	// How much history to keep in memory
  //	Flag indicating whether or not the parameters have been set
  BOOL setData;
  BOOL subPopArrayCreated;	// Whether or not the subpopulation array has
                                // already been created
  BOOL climEconCreated;		// Toggle probability arrays created?
  // GUI Stuff
  id <ButtonPanel> panel;
  BOOL panelCreated;
  StrategyColourKey *strategyKey;
  unsigned seed;		// The seed used for the run
}

+createBegin: (id)z;
-createEnd;
-finalise;
-setEnvironmentType: (char *)value;
-setNeighbourhoodRadius: (unsigned)value;
-setMooreNeighbourhood: (BOOL)value;		// deprecated
-setEnvironmentBounded: (BOOL)value;		// deprecated
-setClimateBitStringSize: (unsigned)value;
-setEconomyBitStringSize: (unsigned)value;
-setParcelBitStringSize: (unsigned)value;
-setNUses: (unsigned)value;
-setPUseDontCareBit: (double)value;
-setNParcelClumpCycles: (unsigned)value;	// deprecated
-setClumper: (id <Clumper>)value;
-setEnvironmentXSize: (unsigned)value;
-setEnvironmentYSize: (unsigned)value;
-setStrategyChangeUnit: (double)value;
-setNeighbourNoiseMax: (double)value;
-setNeighbourNoiseMin: (double)value;
-setBreakEvenThreshold: (double)value;
-setLandParcelPrice: (double)value;
-setNInitXParcels: (unsigned)value;
-setNInitYParcels: (unsigned)value;
-setDumpHistory: (BOOL)value;
-setMaxHistoryLength: (unsigned)value;
-setLandParcelFile: (char *)value;
-setLandUseFile: (char *)value;
-setClimateFile: (char *)value;
-setEconomyFile: (char *)value;
-setClimateToggleProbFile: (char *)value;
-setEconomyToggleProbFile: (char *)value;
-setNSubPopulations: (unsigned)value;
-setSubPopulationFile: (char *)value;
-setMaximumYear: (unsigned)value;
-setInfiniteTime: (BOOL)value;
-setSpecifiedSeed: (unsigned)value;
-setStrategyColourKey: (StrategyColourKey *)key;	// Set from obs swarm
// The "get" methods don't have the word "get", so you can write
// [parameter changeUseThreshold] -- now isn't that nice? No?
-(char *)environmentType;
-(unsigned)neighbourhoodRadius;
-(BOOL)mooreNeighbourhood;	// deprecated
-(BOOL)environmentBounded;
-(unsigned)climateBitStringSize;
-(NumberArray *)climateToggleProbs;
-(unsigned)economyBitStringSize;
-(NumberArray *)economyToggleProbs;
-(unsigned)parcelBitStringSize;
-(unsigned)useBitStringSize;
-(unsigned)nUses;
-(double)pUseDontCareBit;
-(unsigned)nParcelClumpCycles;	// deprecated
-(id <Clumper>)clumper;
-(unsigned)environmentXSize;
-(unsigned)environmentYSize;
-(double)strategyChangeUnit;
-(double)neighbourNoiseMax;
-(double)neighbourNoiseMin;
-(unsigned)maximumMemorySize;
-(double)breakEvenThreshold;
-(double)landParcelPrice;
-(unsigned)nInitXParcels;
-(unsigned)nInitYParcels;
-(BOOL)dumpHistory;
-(unsigned)maxHistoryLength;
-(char *)landParcelFile;
-(BOOL)useLandParcelFile;
-(char *)landUseFile;
-(BOOL)useLandUseFile;
-(char *)climateToggleProbFile;
-(char *)climateFile;
-(BOOL)useClimateFile;
-(char *)economyToggleProbFile;
-(char *)economyFile;
-(BOOL)useEconomyFile;
-(unsigned)nSubPopulations;
-(SubPopulation *)subPopulation: (unsigned)i;
-(BOOL)infiniteTime;
-(unsigned)maximumYear;
-(unsigned)specifiedSeed;
-(StrategyColourKey *)strategyColourKey;		// Called from LPs.
// Load in the parameters
-loadFromFileNamed: (const char *)filename;
// Write out the parameters to the history file
-(void)writeHistory: (FILE *)fp;
// Sub-population extra stuff;
-clickMeToGetSubPopulationArray;	// When you do this, you must have
// selected the number of subpopulations you want. There can be no going back.
-(void)createSubPopulations;		// Create the subpopulations. 
-(void)loadSubPopulations: (BOOL)quitFlag;
-(void)finaliseSubPopulations;
-(void)createClimateEconomyToggleProbs;
-(void)loadClimateEconomyToggleProbs: (BOOL)quitFlag;
-(void)finaliseClimateEconomyToggleProbs;

@end
