/*
    FEARLUS model0-5-1-5: SimpleRandomOptimumCopyingStrategy.m
    Copyright (C) 1999-2004  Macaulay Institute

    This file is part of FEARLUS model0-5-1-5, an agent-based model of land use
    change.

    FEARLUS model0-5-1-5 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-5-1-5 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the SimpleRandomOptimumCopyingStrategy object. 

*/

#import "SimpleRandomOptimumCopyingStrategy.h"
#import "SelectUseBucket.h"
#import "LandUse.h"
#import "LandParcel.h"
#import "LandManager.h"
#import "Environment.h"
#import "Parameter.h"
#import "FearlusArguments.h"

@implementation SimpleRandomOptimumCopyingStrategy

/*

create:withManager:andParameters:

Create the strategy, setting the manager and parameters.

*/

+(id <Strategy>)create: aZone
	   withManager: (LandManager *)lmgr
	 andParameters: (Parameter *)p {
  SimpleRandomOptimumCopyingStrategy *obj;

  obj = [super create: aZone];
  obj->lm = lmgr;
  obj->parameter = p;
  return obj;
}

/*

decideLandUseForParcel:

Decide the land use for the land parcel, using a simple copying
strategy. For each land use, find all its uses by a neighbour of the
land manager in the immediately preceding year, and multiply this by the
neighbourhood weighting of the land manager. Add the result to the
number of uses by the land manager themselves in the immediately
preceding year. Choose at random among those land uses with the highest score.

*/

-(LandUse *)decideLandUseForParcel: (LandParcel *)lp {
  SelectUseBucket *bucket;
  id parcelIndex, parcelList, tempZone;
  LandParcel *myParcel, *neighbouringParcel;
  LandManager *neighbour;
  LandUse *lu;
  int i;

  tempZone = [Zone create: scratchZone];
  bucket = [SelectUseBucket create: tempZone
			    withParameters: parameter
			    andLandUses: [[lp getEnvironment] getLandUses]];
  // Loop through land parcels owned by the land manager
  if([arguments debugCoreMethods]) {
    printf("S: %s: Using simple copying strategy of land manager"
	   " %u to determine land use for land parcel %u at (%d, %d)\n",
	   sel_get_name(_cmd), [lm getPIN], [lp getPIN], [lp getX], [lp getY]);
  }
  for(parcelIndex = [[lm getLandParcels] begin: tempZone], [parcelIndex next];
      [parcelIndex getLoc] == Member;
      [parcelIndex next]) {
    myParcel = [parcelIndex get];
    [bucket addScore: 1.0 toBucketForLandUse: [myParcel getLandUse]];
    if([arguments debugCoreMethods]) {
      printf("S: %s: Land parcel %u at (%d, %d) has land use %u. Score for "
	     "land use incremented by %g\n", sel_get_name(_cmd),
	     [myParcel getPIN], [myParcel getX], [myParcel getY],
	     [[myParcel getLandUse] getPIN], 1.0);
    }
  }
  [parcelIndex drop];
  // Loop through the neighbours, and through their land parcels
  for(i = 0; neighbour = [lm getNextNeighbour: i]; i |= 1) {
    if([arguments debugCoreMethods]) {
      printf("S: %s: Adding scores for land uses of land parcels belonging"
	     " to land manager %u\n", sel_get_name(_cmd), [neighbour getPIN]);
    }
    for(parcelIndex = [[neighbour getLandParcels] begin: tempZone],
	  [parcelIndex next];
	[parcelIndex getLoc] == Member;
	[parcelIndex next]) {
      neighbouringParcel = [parcelIndex get];
      [bucket addScore: [lm getNeighbourWeight]
	      toBucketForLandUse: [neighbouringParcel getLandUse]];
      if([arguments debugCoreMethods]) {
	printf("S: %s: Land parcel %u at (%d, %d) has land use %u. Score of "
	       "land useincremented by %g\n", sel_get_name(_cmd),
	       [neighbouringParcel getPIN], [neighbouringParcel getX],
	       [neighbouringParcel getY],
	       [[neighbouringParcel getLandUse] getPIN],
	       [lm getNeighbourWeight]);
      }
    }
    [parcelIndex drop];
  }
  lu = [bucket getRandomOptimumChoice];
  [bucket drop];
  [tempZone drop];
  return lu;
}

@end
